import os
from io import BytesIO

from PIL import Image as pilim
from django.db.models.fields.files import FieldFile
from reportlab.lib import colors
from reportlab.lib.enums import TA_LEFT, TA_RIGHT, TA_CENTER
from reportlab.lib.pagesizes import A4
from reportlab.lib.styles import getSampleStyleSheet
from reportlab.lib.utils import ImageReader
from reportlab.pdfbase import pdfmetrics
from reportlab.pdfbase.ttfonts import TTFont
from reportlab.platypus import BaseDocTemplate, Image, Spacer
from reportlab.platypus import Frame, PageTemplate
from reportlab.platypus import Paragraph, Table, TableStyle

from cp.mc_utils.user_obj_utils import user_can_view_prices
from doors.mc_utils.pdfcreator import get_pdf_info_translated, get_site_specific_info, get_price_symbol, get_dims_sum, \
    convert_din_direction
from doors.mc_utils.session.parsedoorinfo import parse_door_info
from doors.mc_utils.session.reqprms import RP_DOOR_MODEL, RP_DOOR_LOCK, RP_DOOR_SYSTEM, INITIAL_DOOR_PRICE, \
    RP_DOOR_WIDTH_MM, RP_DOOR_HEIGHT_MM, EXTRA_DOOR_HEIGHT_PRICE, RP_DOOR_DIN, RP_DOOR_TYPE_CODE_NAME, \
    RP_LEFT_SIDE_REPEAT, RP_LEFT_SIDE_WIDTH, LEFT_SIDE_GLASS_PRICE, RP_TOP_GLASS_HEIGHT, TOP_SIDE_GLASS_PRICE, \
    RIGHT_SIDE_GLASS_PRICE, RP_RIGHT_SIDE_WIDTH, RP_RIGHT_SIDE_REPEAT, SIDE_GLASSES_BASE_PRICE, SIDE_GLASSES_TOTAL_PRICE
from doors.models import DoorModel
from orders.views import get_prices_of_order
from prices.views import calc_total_price

styles = getSampleStyleSheet()
styleN = styles['Normal']
styleH = styles['Heading1']

DESIRED_PHOTO_WIDTH = A4[0] - 100
DESIRED_PHOTO_HEIGHT = A4[1] / 2 - 170

BASIC_MARGIN = 30
LEFT_MARGIN = BASIC_MARGIN
RIGHT_MARGIN = BASIC_MARGIN
TOP_MARGIN = BASIC_MARGIN
BOTTOM_MARGIN = BASIC_MARGIN
BODY_FONT_SIZE = 8
PAGE_WIDTH = 530
PAGE_HEIGHT = 750
LINE_Y = 730
BODY_STYLE = None
USER_CAN_VIEW_PRICES = None

WATERMARK = None
LOGO_PATH = None
__location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))


class MCPDFBuilder:

    def __init__(self, logo_path=None, watermark=None):
        global WATERMARK
        global LOGO_PATH
        global BODY_STYLE

        LOGO_PATH = logo_path
        WATERMARK = watermark
        BODY_STYLE = get_body_style()

        self.pdf_buffer = BytesIO()
        self.pdf = BaseDocTemplate(self.pdf_buffer, pagesize=A4)
        frame = Frame(LEFT_MARGIN, BOTTOM_MARGIN, PAGE_WIDTH, 687, showBoundary=1)
        template = PageTemplate(id='all_pages', frames=frame, onPage=header_and_footer)
        self.pdf.addPageTemplates([template])
        self.story = []

        # images_before_table = []
        # images_after_table = []
        # if images_data:
        #     for img_path, img_pos in images_data.items():
        #         if img_pos == IMAGE_DOC_POSITION_BEFORE_TABLE:
        #             images_before_table.append(img_path)
        #         elif img_pos == IMAGE_DOC_POSITION_AFTER_TABLE:
        #             images_after_table.append(img_path)
        #
        #
        # for img_path in images_before_table:
        #     self.story.extend([
        #         Spacer(10, 10),
        #         Image(img_path, width=100, height=100),
        #         Spacer(10, 10)
        #     ])
        #
        # self.story.append(pfd_table_builder(data))
        #
        # for img_path in images_after_table:
        #     self.story.extend([
        #         Spacer(10, 10),
        #         Image(img_path, width=100, height=100),
        #         Spacer(10, 10)
        #     ])

    def add_to_story(self, obj):
        if isinstance(obj, list):
            for x in obj:
                self.story.append(x)
        else:
            self.story.append(obj)

    def build(self):
        self.pdf.build(self.story)
        pdf_value = self.pdf_buffer.getvalue()
        self.pdf_buffer.close()
        return pdf_value


class Row:
    def __init__(self, columns, grid=None, background_color=None, text_color=None, full_width=False, align=TA_LEFT):
        self.columns = columns
        self.grid = grid
        self.background_color = background_color
        self.text_color = text_color
        self.align = align
        self.full_width = full_width

    def get_number_of_columns(self):
        return len(self.columns)


class Column:

    def __init__(self, content, grid=None, background_color=None, text_color=None, image_width=50, image_height=50,
                 align=TA_LEFT):
        """

        :param content: This variable's data will be passed into the Reportlab's Paragraph Class, or if its BytesIO
                        object will be passed in a Reportlab's Image Class object and appended in the table.
        :param grid:
        :param background_color:
        :param text_color:
        :param image_width:
        :param image_height:
        :param align: Align options (from ReportLab) -> TA_LEFT, TA_CENTER, TA_RIGHT, TA_JUSTIFY
        """
        self.content = content
        self.grid = grid
        self.background_color = background_color
        self.text_color = text_color
        self.image_width = image_width
        self.image_height = image_height
        self.align = align


class Grid:
    def __init__(self, line_width=0.5, line_color=colors.gray):
        self.line_width = line_width
        self.line_color = line_color


def get_img(img_path):
    # buff1 = BytesIO()
    # img = pilim.open(img_path)
    # print('img_path:', img_path, img.size[0], img.size[1])
    # door_out_img = Image(img_path, width=2 * inch, height=2 * inch)

    return '''<para autoLeading="off" fontSize=12>This &lt;img/&gt; <img
src="{0}" valign="top" width="{1}" height="{2}"/> </para>'''.format(img_path, 100, 100)


def pfd_table_builder(data):
    """
    Example usage:

    data = [
        Row(columns=[Column(content='<font size=10>{}</font>'.format(datetime.date.today().strftime("%d/%m/%Y")))]),
        Spacer(1, 1),
        Row(columns=[Column(content='Title')]),
        Spacer(1, 1),
        Image(buff1, width=150, height=300),
        Row(
            columns=[
                Column(content='Panel color outside:'),
                Column(content='7012 Matt'),
                Column(content='$100', align=TA_RIGHT),
                Column(content='test'),
                Column(content='test2', grid=Grid(line_color=colors.yellow, line_width=2)),
            ],
            grid=Grid(line_color=colors.blue, line_width=grid_width),
            # text_color=colors.red
        ),
        Row(
            columns=[
                Column(content='test2', align=TA_CENTER, background_color=colors.red, text_color=colors.blue),
                Column(content='Panel color outside:'),
                Column(content='7012 Matt'),
                Column(content='$100', align=TA_RIGHT, grid=Grid(line_color=colors.yellow, line_width=2)),
                Column(content='test', align=TA_CENTER, background_color=colors.aqua),
                Column(content='test2', align=TA_CENTER, background_color=colors.red, text_color=colors.white),
                Column(content='test2', align=TA_CENTER, background_color=colors.purple, text_color=colors.red),

            ],
            align=TA_LEFT,
            background_color=colors.green,
            text_color=colors.red,
        ),
        Row(
            columns=[
                Column(content='test2', align=TA_CENTER, background_color=colors.red, text_color=colors.blue),
                Column(content='Panel color outside:'),
                Column(content='7012 Matt'),
                Column(content='$100', align=TA_RIGHT),
                Column(content='test1', background_color=colors.purple),
                Column(content='test', align=TA_CENTER, background_color=colors.aqua, text_color=colors.yellow),
                Column(content='test2', align=TA_CENTER, background_color=colors.red, text_color=colors.white),
            ],
            background_color=colors.green,
            text_color=colors.blue,
            grid=Grid(),
        ),
    ]

    table = pfd_table_builder(data)

    :param data:
    :return: A reportlab table instance
    """

    table_data = []
    for row_ctr, row_obj in enumerate(data):
        if isinstance(row_obj, Row):
            column_final = []
            for col_ctr, col in enumerate(row_obj.columns):
                style = get_body_style()
                # print('{}, {}'.format(row_ctr, col_ctr))
                # Loop through all the cells and apply the styles on cell level if declared
                # specifically for the particular cell or else apply the style of the row on the cell (if any).
                # Align
                if col.align:
                    style.alignment = col.align
                elif row_obj.align:
                    style.alignment = row_obj.align
                # Text Colour
                if col.text_color:
                    style.textColor = col.text_color
                elif row_obj.text_color:
                    style.textColor = row_obj.text_color

                # print("type:", type(col.content))
                if isinstance(col.content, BytesIO) or isinstance(col.content, FieldFile):
                    _col = Image(col.content, width=col.image_width, height=col.image_height)
                elif isinstance(col.content, str):
                    _col = Paragraph('{}'.format(col.content, 'N/A'), style)
                else:
                    # print("SOMETHING ELSE")
                    _col = col.content

                column_final.append(_col)

            table_data.append(column_final)
        else:
            table_data.append([row_obj])

    # ------------------ STYLE
    table_styles = TableStyle()
    for row_ctr, row_obj in enumerate(data):
        if isinstance(row_obj, Row):
            # Loop through all the cells and apply the styles on cell level if declared
            # specifically for the particular cell or else apply the style of the row on the cell
            # (so every cell on this row will have this style, if any).
            for col_ctr, col_obj in enumerate(row_obj.columns):
                apply_style = False
                # Background Colour
                background_color = None
                if col_obj.background_color:
                    apply_style = True
                    background_color = col_obj.background_color
                elif row_obj.background_color:
                    apply_style = True
                    background_color = row_obj.background_color
                if apply_style:
                    table_styles.add('BACKGROUND', (col_ctr, row_ctr), (col_ctr, row_ctr), background_color)
                # Grid
                apply_style = False
                grid_line_color = None
                grid_line_width = None
                if col_obj.grid:
                    apply_style = True
                    grid_line_color = col_obj.grid.line_color
                    grid_line_width = col_obj.grid.line_width
                elif row_obj.grid:
                    apply_style = True
                    grid_line_color = row_obj.grid.line_color
                    grid_line_width = row_obj.grid.line_width
                if apply_style:
                    table_styles.add('GRID', (col_ctr, row_ctr), (col_ctr, row_ctr), grid_line_width, grid_line_color)
                # SPAN
                if row_obj.full_width:
                    table_styles.add('SPAN', (col_ctr, row_ctr), (-1, row_ctr))

    t = Table(table_data)
    t.setStyle(table_styles)
    return t


def header_and_footer(canvas, pdf):
    # LOGO
    # print('LOGO_PATH', LOGO_PATH)
    if LOGO_PATH:
        im = pilim.open(LOGO_PATH)
        pil_img = ImageReader(im)
        canvas.drawImage(pil_img, 350, PAGE_HEIGHT, width=210, height=70, mask='auto')

    canvas.line(BASIC_MARGIN, LINE_Y, PAGE_WIDTH + BASIC_MARGIN, LINE_Y)

    # ----------------------- FOOTER ----------------------- #
    canvas.setFont('Helvetica-Bold', 8)
    # page number
    page_number_text = "%d" % pdf.page
    canvas.drawCentredString(295, 8, '-' + page_number_text + '-')
    # watermark
    if WATERMARK:
        canvas.drawCentredString(507, 8, WATERMARK)

    canvas.line(BASIC_MARGIN, 20, PAGE_WIDTH + BASIC_MARGIN, 20)


def get_body_style():
    sample_style_sheet = getSampleStyleSheet()
    body_style = sample_style_sheet['BodyText']
    body_style.fontSize = BODY_FONT_SIZE
    pdfmetrics.registerFont(TTFont('NotoSans', os.path.join(__location__, 'fonts/NotoSans-Regular.ttf')))
    pdfmetrics.registerFont(TTFont('NotoSansBold', os.path.join(__location__, 'fonts/NotoSans-Bold.ttf')))
    body_style.fontName = 'NotoSans'
    return body_style


def find_max_num_of_columns(data):
    max_num_of_cols = 0
    for row_counter, row in enumerate(data):
        num_of_cols = len(row.columns) if isinstance(row, Row) else 1
        if num_of_cols > max_num_of_cols:
            max_num_of_cols = num_of_cols
    return max_num_of_cols


def append_colon(param):
    return '{}:'.format(param)


def get_pdf_2(di_values, price_info_dict, door_img_out_buff, door_img_in_buff, door_img_size, client, order):
    ttt = get_pdf_info_translated()
    door_model = DoorModel.objects.get(pk=di_values.get(RP_DOOR_MODEL))
    di_names = parse_door_info(di_values, True, door_model)
    door_type = di_names.get(RP_DOOR_TYPE_CODE_NAME)
    we_have_side_glasses = 'l' in door_type or 'r' in door_type or 't' in door_type
    grid_width = 0.5
    grid_color = colors.black
    calculated_prices = None
    total_price = None
    is_order = False
    global USER_CAN_VIEW_PRICES
    USER_IS_CLIENT = user_can_view_prices(client)

    if order:
        is_order = True
    if USER_IS_CLIENT:
        if is_order:
            total_price, calculated_prices = get_prices_of_order(order)
        else:
            total_price, calculated_prices = calc_total_price(
                di_values,
                price_info_dict,
                client,
                is_pdf=True
            )
    mc_colors = {
        'grey_color': '#cccccc',
        'light_blue_color': '#cfe7f5',
        'money_cell_color': '#ffffff',
        'title_color': '#999999',
        'images_subtitle': '#004586',
    }

    # ---------------------------------------------------------------------------------------------------------------- #
    # LEFT TABLE ---------------------------------------------------------------------------------------------------- #
    # ---------------------------------------------------------------------------------------------------------------- #
    left_table_data = [
        # Door Model
        Row(
            columns=[
                Column(content=append_colon(ttt[1]), background_color=mc_colors['grey_color']),
                Column(content=get_site_specific_info(door_model, di_values, di_names)),
                Column(
                    content="{}{}".format(
                        calculated_prices.get(INITIAL_DOOR_PRICE), get_price_symbol()
                    ) if USER_IS_CLIENT else '',
                    background_color=mc_colors['grey_color'],
                    align=TA_RIGHT
                ),
            ],
            grid=Grid(line_color=grid_color, line_width=grid_width),
        ),

        # System
        Row(
            columns=[
                Column(content=append_colon(ttt[5]), background_color=mc_colors['grey_color']),
                Column(content=di_names.get(RP_DOOR_SYSTEM)[0] + ' ' + di_names.get(RP_DOOR_SYSTEM)[1]),
                Column(content=''),
            ],
            grid=Grid(line_color=grid_color, line_width=grid_width),
        ),
        # Κάσα
        Row(
            columns=[
                Column(content=append_colon('Κάσα'), background_color=mc_colors['grey_color']),
                Column(content=''),
                Column(content='', background_color=mc_colors['money_cell_color'], align=TA_RIGHT),
            ],
            grid=Grid(line_color=grid_color, line_width=grid_width),
        ),
        #  Overall Dimensions
        Row(
            columns=[
                Column(content=append_colon(ttt[6]), background_color=mc_colors['light_blue_color']),
                Column(content="{} mm".format(get_dims_sum(di_names))),
                Column(content='', background_color=mc_colors['money_cell_color'], align=TA_RIGHT),
            ],
            grid=Grid(line_color=grid_color, line_width=grid_width),
        ),
        #  Door Dimensions
        Row(
            columns=[
                Column(content=append_colon(ttt[7]), background_color=mc_colors['light_blue_color']),
                Column(content="{} x {} mm".format(
                    di_names.get(RP_DOOR_WIDTH_MM),
                    di_names.get(RP_DOOR_HEIGHT_MM)
                )),
                Column(
                    content=(
                        "({}) {}{}".format(
                            ttt[66],
                            round(calculated_prices.get(EXTRA_DOOR_HEIGHT_PRICE), 2),
                            get_price_symbol()
                        ) if calculated_prices.get(EXTRA_DOOR_HEIGHT_PRICE) > 0 else ""
                    ) if USER_IS_CLIENT else '',
                    background_color=mc_colors['money_cell_color'],
                    align=TA_RIGHT
                ),
            ],
            grid=Grid(line_color=grid_color, line_width=grid_width),
        ),
        #  Din Direction
        Row(
            columns=[
                Column(content=append_colon(ttt[14]), background_color=mc_colors['light_blue_color']),
                Column(content=convert_din_direction(di_names.get(RP_DOOR_DIN), ttt)),
                Column(content='', background_color=mc_colors['money_cell_color'], align=TA_RIGHT),
            ],
            grid=Grid(line_color=grid_color, line_width=grid_width),
        ),
    ]
    # Base Sidelight Glass Price
    if USER_IS_CLIENT and we_have_side_glasses:
        left_table_data += [
            Row(
                columns=[
                    Column(content=append_colon(ttt[70]), background_color=mc_colors['grey_color']),
                    Column(content=''),
                    Column(
                        content="{}{}".format(
                            round(calculated_prices.get(SIDE_GLASSES_BASE_PRICE), 2),
                            get_price_symbol()
                        ) if USER_IS_CLIENT else '',
                        background_color=mc_colors['money_cell_color'],
                        align=TA_RIGHT
                    ),
                ],
                grid=Grid(line_color=grid_color, line_width=grid_width),
            ),
        ]
    # Left Sidelight Glass
    if 'l' in door_type:
        left_table_data += [
            Row(
                columns=[
                    Column(content=append_colon(ttt[9]), background_color=mc_colors['grey_color']),
                    Column(
                        content='{} mm'.format(di_names.get(RP_LEFT_SIDE_WIDTH)) if str(di_names.get(
                            RP_LEFT_SIDE_REPEAT)) == '1' else '{} mm (x{})'.format(
                            di_names.get(RP_LEFT_SIDE_WIDTH),
                            di_names.get(RP_LEFT_SIDE_REPEAT)
                        )),
                    Column(
                        content="{}{}".format(
                            round(calculated_prices.get(LEFT_SIDE_GLASS_PRICE), 2),
                            get_price_symbol()
                        ) if USER_IS_CLIENT else '',
                        background_color=mc_colors['money_cell_color'],
                        align=TA_RIGHT
                    ),
                ],
                grid=Grid(line_color=grid_color, line_width=grid_width),
            ),
        ]
    # Right Sidelight Glass
    if 'r' in door_type:
        left_table_data += [
            Row(
                columns=[
                    Column(content=append_colon(ttt[10]), background_color=mc_colors['grey_color']),
                    Column(
                        content='{} mm'.format(di_names.get(RP_RIGHT_SIDE_WIDTH)) if str(di_names.get(
                            RP_RIGHT_SIDE_REPEAT)) == '1' else '{} mm (x{})'.format(
                            di_names.get(RP_RIGHT_SIDE_WIDTH),
                            di_names.get(RP_RIGHT_SIDE_REPEAT)
                        )
                    ),
                    Column(
                        content="{}{}".format(
                            round(calculated_prices.get(RIGHT_SIDE_GLASS_PRICE), 2),
                            get_price_symbol()
                        ) if USER_IS_CLIENT else '',
                        background_color=mc_colors['money_cell_color'],
                        align=TA_RIGHT
                    ),
                ],
                grid=Grid(line_color=grid_color, line_width=grid_width),
            ),
        ]
    # Top Sidelight Glass
    if 't' in door_type:
        left_table_data += [
            Row(
                columns=[
                    Column(content=append_colon(ttt[11]), background_color=mc_colors['grey_color']),
                    Column(content=str(di_names.get(RP_TOP_GLASS_HEIGHT)) + ' mm'),
                    Column(
                        content="{}{}".format(
                            round(calculated_prices.get(TOP_SIDE_GLASS_PRICE), 2),
                            get_price_symbol()
                        ) if USER_IS_CLIENT else '',
                        background_color=mc_colors['money_cell_color'],
                        align=TA_RIGHT
                    ),
                ],
                grid=Grid(line_color=grid_color, line_width=grid_width),
            ),
        ]
    # Total Sidelight Glasses price
    if USER_IS_CLIENT and we_have_side_glasses:
        left_table_data += [
            Row(
                columns=[
                    Column(content=append_colon(ttt[72]), background_color=mc_colors['grey_color']),
                    Column(content=''),
                    Column(
                        content="{}{}".format(
                            round(calculated_prices.get(SIDE_GLASSES_TOTAL_PRICE), 2), get_price_symbol()
                        ),
                        background_color=mc_colors['money_cell_color'],
                        align=TA_RIGHT
                    ),
                ],
                grid=Grid(line_color=grid_color, line_width=grid_width),
            ),
        ]
    # Colours
    left_table_data += [
        # Panel Colour Out
        Row(
            columns=[
                Column(content=append_colon(ttt[73]), background_color=mc_colors['light_blue_color']),
                Column(content=''),
                Column(
                    content='',
                    background_color=mc_colors['money_cell_color'],
                    align=TA_RIGHT
                ),
            ],
            grid=Grid(line_color=grid_color, line_width=grid_width),
        ),
        # Door Frame Colour Out (Χρώμα  Κάσας Έξω)
        Row(
            columns=[
                Column(content=append_colon(ttt[74]), background_color=mc_colors['light_blue_color']),
                Column(content=''),
                Column(
                    content='',
                    background_color=mc_colors['money_cell_color'],
                    align=TA_RIGHT
                ),
            ],
            grid=Grid(line_color=grid_color, line_width=grid_width),
        ),
        # Frame Colour Out (Χρώμα Κορνίζας Έξω)
        Row(
            columns=[
                Column(content=append_colon(ttt[75]), background_color=mc_colors['light_blue_color']),
                Column(content=''),
                Column(
                    content='',
                    background_color=mc_colors['money_cell_color'],
                    align=TA_RIGHT
                ),
            ],
            grid=Grid(line_color=grid_color, line_width=grid_width),
        ),
        # Secondary Panel Colour
        Row(
            columns=[
                Column(content=append_colon(ttt[79]), background_color=mc_colors['light_blue_color']),
                Column(content=''),
                Column(
                    content='',
                    background_color=mc_colors['money_cell_color'],
                    align=TA_RIGHT
                ),
            ],
            grid=Grid(line_color=grid_color, line_width=grid_width),
        ),
        # Panel Colour In
        Row(
            columns=[
                Column(content=append_colon(ttt[76]), background_color=mc_colors['light_blue_color']),
                Column(content=''),
                Column(
                    content='',
                    background_color=mc_colors['money_cell_color'],
                    align=TA_RIGHT
                ),
            ],
            grid=Grid(line_color=grid_color, line_width=grid_width),
        ),
        # Door Frame Colour In (Χρώμα Κάσας Μέσα)
        Row(
            columns=[
                Column(content=append_colon(ttt[77]), background_color=mc_colors['light_blue_color']),
                Column(content=''),
                Column(
                    content='',
                    background_color=mc_colors['money_cell_color'],
                    align=TA_RIGHT
                ),
            ],
            grid=Grid(line_color=grid_color, line_width=grid_width),
        ),
        # Frame Colour In (Χρώμα Κορνίζας Μέσα)
        Row(
            columns=[
                Column(content=append_colon(ttt[78]), background_color=mc_colors['light_blue_color']),
                Column(content=''),
                Column(
                    content='',
                    background_color=mc_colors['money_cell_color'],
                    align=TA_RIGHT
                ),
            ],
            grid=Grid(line_color=grid_color, line_width=grid_width),
        ),
    ]

    # ---------------------------------------------------------------------------------------------------------------- #
    # RIGHT TABLE ---------------------------------------------------------------------------------------------------- #
    # ---------------------------------------------------------------------------------------------------------------- #
    right_table = [
        # Lock
        Row(
            columns=[
                Column(content=append_colon(ttt[43]), background_color=mc_colors['light_blue_color']),
                Column(content=di_names.get(RP_DOOR_LOCK)),
                Column(
                    content="{}{}".format(price_info_dict.get(RP_DOOR_LOCK), get_price_symbol()),
                    background_color=mc_colors['money_cell_color'],
                    align=TA_RIGHT
                ),
            ],
            grid=Grid(line_color=grid_color, line_width=grid_width),
        ),
        # Προέκταση
        Row(
            columns=[
                Column(content=append_colon('Προέκταση'), background_color=mc_colors['light_blue_color']),
                Column(content=''),
                Column(content='$100', background_color=mc_colors['money_cell_color'], align=TA_RIGHT),
            ],
            grid=Grid(line_color=grid_color, line_width=grid_width),
        ),
    ]

    # ---------------------------------------------------------------------------------------------------------------- #
    # What's Included (Βασικές παροχές) ------------------------------------------------------------------------------ #
    # ---------------------------------------------------------------------------------------------------------------- #
    wats_included_table = [
        # Title
        Row(
            columns=[
                Column(
                    content=append_colon(ttt[80]),
                    align=TA_CENTER,
                    text_color=colors.white,
                    background_color=mc_colors['title_color']
                ),
            ],
            grid=Grid(line_color=grid_color, line_width=grid_width),
        ),
        # Images
        Row(
            columns=[
                Column(
                    content=door_img_in_buff,
                    align=TA_CENTER,
                    background_color=mc_colors['grey_color'],
                ),
                Column(
                    content=door_img_in_buff,
                    align=TA_CENTER,
                    background_color=mc_colors['grey_color'],
                ),
                Column(
                    content=door_img_in_buff,
                    align=TA_CENTER,
                    background_color=mc_colors['grey_color'],
                ),
                Column(
                    content=door_img_in_buff,
                    align=TA_CENTER,
                    background_color=mc_colors['grey_color'],
                ), Column(
                    content=door_img_in_buff,
                    align=TA_CENTER,
                    background_color=mc_colors['grey_color'],
                ),
                Column(
                    content=door_img_in_buff,
                    align=TA_CENTER,
                    background_color=mc_colors['grey_color'],
                ),

            ],
            grid=Grid(line_color=grid_color, line_width=grid_width),
        ),
        # Image Subtitles
        Row(
            columns=[
                Column(
                    content='Image Subtitle',
                    align=TA_CENTER,
                    text_color=mc_colors['images_subtitle'],
                    background_color=mc_colors['grey_color'],
                ),
                Column(
                    content='Image Subtitle',
                    align=TA_CENTER,
                    text_color=mc_colors['images_subtitle'],
                    background_color=mc_colors['grey_color'],
                ),
                Column(
                    content='Image Subtitle',
                    align=TA_CENTER,
                    text_color=mc_colors['images_subtitle'],
                    background_color=mc_colors['grey_color'],
                ),
                Column(
                    content='Image Subtitle',
                    align=TA_CENTER,
                    text_color=mc_colors['images_subtitle'],
                    background_color=mc_colors['grey_color'],
                ),
                Column(
                    content='Image Subtitle',
                    align=TA_CENTER,
                    text_color=mc_colors['images_subtitle'],
                    background_color=mc_colors['grey_color'],
                ),
                Column(
                    content='Image Subtitle',
                    align=TA_CENTER,
                    text_color=mc_colors['images_subtitle'],
                    background_color=mc_colors['grey_color'],
                ),
            ],
            grid=Grid(line_color=grid_color, line_width=grid_width),
        ),
    ]

    # ---------------------------------------------------------------------------------------------------------------- #
    # BUILD PDF ------------------------------------------------------------------------------------------------------ #
    # ---------------------------------------------------------------------------------------------------------------- #
    left_table = pfd_table_builder(left_table_data)
    right_table = pfd_table_builder(right_table)
    two_vertical_tables = Table([(left_table, right_table)])
    two_vertical_tables.setStyle(TableStyle([
        ('VALIGN', (0, 0), (-1, -1), 'TOP')
    ]))

    whats_included = pfd_table_builder(wats_included_table)

    mc_pdf_builder = MCPDFBuilder()
    mc_pdf_builder.add_to_story(two_vertical_tables)
    mc_pdf_builder.add_to_story(Spacer(1, 50))
    mc_pdf_builder.add_to_story(whats_included)
    return mc_pdf_builder.build()
